@@ -1,8 +1,8 @@
 use crate::{
     front::glsl::{
         ast::{
-            GlobalLookup, GlobalLookupKind, Precision, StorageQualifier, StructLayout,
-            TypeQualifier,
+            GlobalLookup, GlobalLookupKind, Precision, QualifierKey, QualifierValue,
+            StorageQualifier, StructLayout, TypeQualifiers,
         },
         context::{Context, ExprPos},
         error::ExpectedToken,
@@ -10,30 +10,79 @@ use crate::{
         token::{Token, TokenValue},
         types::scalar_components,
         variables::{GlobalOrConstant, VarDeclaration},
-        Error, ErrorKind, Parser, Span,
+        Error, ErrorKind, Frontend, Span,
     },
-    Block, Expression, FunctionResult, Handle, ScalarKind, Statement, StorageClass, StructMember,
-    Type, TypeInner,
+    proc::Alignment,
+    AddressSpace, Expression, FunctionResult, Handle, ScalarKind, Statement, StructMember, Type,
+    TypeInner,
 };
 
 use super::{DeclarationContext, ParsingContext, Result};
 
+/// Helper method used to retrieve the child type of `ty` at
+/// index `i`.
+///
+/// # Note
+///
+/// Does not check if the index is valid and returns the same type
+/// when indexing out-of-bounds a struct or indexing a non indexable
+/// type.
+fn element_or_member_type(
+    ty: Handle<Type>,
+    i: usize,
+    types: &mut crate::UniqueArena<Type>,
+) -> Handle<Type> {
+    match types[ty].inner {
+        // The child type of a vector is a scalar of the same kind and width
+        TypeInner::Vector { kind, width, .. } => types.insert(
+            Type {
+                name: None,
+                inner: TypeInner::Scalar { kind, width },
+            },
+            Default::default(),
+        ),
+        // The child type of a matrix is a vector of floats with the same
+        // width and the size of the matrix rows.
+        TypeInner::Matrix { rows, width, .. } => types.insert(
+            Type {
+                name: None,
+                inner: TypeInner::Vector {
+                    size: rows,
+                    kind: ScalarKind::Float,
+                    width,
+                },
+            },
+            Default::default(),
+        ),
+        // The child type of an array is the base type of the array
+        TypeInner::Array { base, .. } => base,
+        // The child type of a struct at index `i` is the type of it's
+        // member at that same index.
+        //
+        // In case the index is out of bounds the same type is returned
+        TypeInner::Struct { ref members, .. } => {
+            members.get(i).map(|member| member.ty).unwrap_or(ty)
+        }
+        // The type isn't indexable, the same type is returned
+        _ => ty,
+    }
+}
+
 impl<'source> ParsingContext<'source> {
     pub fn parse_external_declaration(
         &mut self,
-        parser: &mut Parser,
+        frontend: &mut Frontend,
         global_ctx: &mut Context,
-        global_body: &mut Block,
     ) -> Result<()> {
         if self
-            .parse_declaration(parser, global_ctx, global_body, true)?
+            .parse_declaration(frontend, global_ctx, true, false)?
             .is_none()
         {
-            let token = self.bump(parser)?;
+            let token = self.bump(frontend)?;
             match token.value {
-                TokenValue::Semicolon if parser.meta.version == 460 => Ok(()),
+                TokenValue::Semicolon if frontend.meta.version == 460 => Ok(()),
                 _ => {
-                    let expected = match parser.meta.version {
+                    let expected = match frontend.meta.version {
                         460 => vec![TokenValue::Semicolon.into(), ExpectedToken::Eof],
                         _ => vec![ExpectedToken::Eof],
                     };
@@ -50,10 +99,9 @@ impl<'source> ParsingContext<'source> {
 
     pub fn parse_initializer(
         &mut self,
-        parser: &mut Parser,
+        frontend: &mut Frontend,
         ty: Handle<Type>,
         ctx: &mut Context,
-        body: &mut Block,
     ) -> Result<(Handle<Expression>, Span)> {
         // initializer:
         //     assignment_expression
@@ -63,18 +111,20 @@ impl<'source> ParsingContext<'source> {
         // initializer_list:
         //     initializer
         //     initializer_list COMMA initializer
-        if let Some(Token { mut meta, .. }) = self.bump_if(parser, TokenValue::LeftBrace) {
+        if let Some(Token { mut meta, .. }) = self.bump_if(frontend, TokenValue::LeftBrace) {
             // initializer_list
             let mut components = Vec::new();
             loop {
-                // TODO: Change type
-                components.push(self.parse_initializer(parser, ty, ctx, body)?.0);
+                // The type expected to be parsed inside the initializer list
+                let new_ty = element_or_member_type(ty, components.len(), &mut ctx.module.types);
 
-                let token = self.bump(parser)?;
+                components.push(self.parse_initializer(frontend, new_ty, ctx)?.0);
+
+                let token = self.bump(frontend)?;
                 match token.value {
                     TokenValue::Comma => {
                         if let Some(Token { meta: end_meta, .. }) =
-                            self.bump_if(parser, TokenValue::RightBrace)
+                            self.bump_if(frontend, TokenValue::RightBrace)
                         {
                             meta.subsume(end_meta);
                             break;
@@ -97,17 +147,17 @@ impl<'source> ParsingContext<'source> {
             }
 
             Ok((
-                ctx.add_expression(Expression::Compose { ty, components }, meta, body),
+                ctx.add_expression(Expression::Compose { ty, components }, meta)?,
                 meta,
             ))
         } else {
             let mut stmt = ctx.stmt_ctx();
-            let expr = self.parse_assignment(parser, ctx, &mut stmt, body)?;
-            let (mut init, init_meta) = ctx.lower_expect(stmt, parser, expr, ExprPos::Rhs, body)?;
+            let expr = self.parse_assignment(frontend, ctx, &mut stmt)?;
+            let (mut init, init_meta) = ctx.lower_expect(stmt, frontend, expr, ExprPos::Rhs)?;
 
-            let scalar_components = scalar_components(&parser.module.types[ty].inner);
+            let scalar_components = scalar_components(&ctx.module.types[ty].inner);
             if let Some((kind, width)) = scalar_components {
-                ctx.implicit_conversion(parser, &mut init, init_meta, kind, width)?;
+                ctx.implicit_conversion(&mut init, init_meta, kind, width)?;
             }
 
             Ok((init, init_meta))
@@ -119,9 +169,8 @@ impl<'source> ParsingContext<'source> {
     // produced Error::InvalidToken if it isn't consumed
     pub fn parse_init_declarator_list(
         &mut self,
-        parser: &mut Parser,
-        ty: Handle<Type>,
-        mut fallthrough: Option<Token>,
+        frontend: &mut Frontend,
+        mut ty: Handle<Type>,
         ctx: &mut DeclarationContext,
     ) -> Result<()> {
         // init_declarator_list:
@@ -139,20 +188,15 @@ impl<'source> ParsingContext<'source> {
         //     fully_specified_type IDENTIFIER EQUAL initializer
 
         // Consume any leading comma, e.g. this is valid: `float, a=1;`
-        if fallthrough
-            .as_ref()
-            .or_else(|| self.peek(parser))
-            .filter(|t| t.value == TokenValue::Comma)
-            .is_some()
+        if self
+            .peek(frontend)
+            .map_or(false, |t| t.value == TokenValue::Comma)
         {
-            fallthrough.take().or_else(|| self.next(parser));
+            self.next(frontend);
         }
 
         loop {
-            let token = fallthrough
-                .take()
-                .ok_or(ErrorKind::EndOfFile)
-                .or_else(|_| self.bump(parser))?;
+            let token = self.bump(frontend)?;
             let name = match token.value {
                 TokenValue::Semicolon => break,
                 TokenValue::Identifier(name) => name,
@@ -175,41 +219,63 @@ impl<'source> ParsingContext<'source> {
             // parse an array specifier if it exists
             // NOTE: unlike other parse methods this one doesn't expect an array specifier and
             // returns Ok(None) rather than an error if there is not one
-            let array_specifier = self.parse_array_specifier(parser)?;
-            let ty = parser.maybe_array(ty, meta, array_specifier);
+            self.parse_array_specifier(frontend, ctx.ctx, &mut meta, &mut ty)?;
+
+            let is_global_const =
+                ctx.qualifiers.storage.0 == StorageQualifier::Const && ctx.external;
 
             let init = self
-                .bump_if(parser, TokenValue::Assign)
+                .bump_if(frontend, TokenValue::Assign)
                 .map::<Result<_>, _>(|_| {
-                    let (mut expr, init_meta) =
-                        self.parse_initializer(parser, ty, ctx.ctx, ctx.body)?;
+                    let prev_const = ctx.ctx.is_const;
+                    ctx.ctx.is_const = is_global_const;
 
-                    let scalar_components = scalar_components(&parser.module.types[ty].inner);
+                    let (mut expr, init_meta) = self.parse_initializer(frontend, ty, ctx.ctx)?;
+
+                    let scalar_components = scalar_components(&ctx.ctx.module.types[ty].inner);
                     if let Some((kind, width)) = scalar_components {
                         ctx.ctx
-                            .implicit_conversion(parser, &mut expr, init_meta, kind, width)?;
+                            .implicit_conversion(&mut expr, init_meta, kind, width)?;
                     }
 
+                    ctx.ctx.is_const = prev_const;
+
                     meta.subsume(init_meta);
 
-                    Ok((expr, init_meta))
+                    Ok(expr)
                 })
                 .transpose()?;
 
-            // TODO: Should we try to make constants here?
-            // This is mostly a hack because we don't yet support adding
-            // bodies to entry points for variable initialization
-            let maybe_constant =
-                init.and_then(|(root, meta)| parser.solve_constant(ctx.ctx, root, meta).ok());
+            let decl_initializer;
+            let late_initializer;
+            if is_global_const {
+                decl_initializer = init;
+                late_initializer = None;
+            } else if ctx.external {
+                decl_initializer =
+                    init.and_then(|expr| ctx.ctx.lift_up_const_expression(expr).ok());
+                late_initializer = None;
+            } else if let Some(init) = init {
+                if ctx.is_inside_loop || !ctx.ctx.expression_constness.is_const(init) {
+                    decl_initializer = None;
+                    late_initializer = Some(init);
+                } else {
+                    decl_initializer = Some(init);
+                    late_initializer = None;
+                }
+            } else {
+                decl_initializer = None;
+                late_initializer = None;
+            };
 
-            let pointer = ctx.add_var(parser, ty, name, maybe_constant, meta)?;
+            let pointer = ctx.add_var(frontend, ty, name, decl_initializer, meta)?;
 
-            if let Some((value, _)) = init.filter(|_| maybe_constant.is_none()) {
-                ctx.flush_expressions();
-                ctx.body.push(Statement::Store { pointer, value }, meta);
+            if let Some(value) = late_initializer {
+                ctx.ctx.emit_restart();
+                ctx.ctx.body.push(Statement::Store { pointer, value }, meta);
             }
 
-            let token = self.bump(parser)?;
+            let token = self.bump(frontend)?;
             match token.value {
                 TokenValue::Semicolon => break,
                 TokenValue::Comma => {}
@@ -231,10 +297,10 @@ impl<'source> ParsingContext<'source> {
     /// `external` whether or not we are in a global or local context
     pub fn parse_declaration(
         &mut self,
-        parser: &mut Parser,
+        frontend: &mut Frontend,
         ctx: &mut Context,
-        body: &mut Block,
         external: bool,
+        is_inside_loop: bool,
     ) -> Result<Option<Span>> {
         //declaration:
         //    function_prototype  SEMICOLON
@@ -248,36 +314,35 @@ impl<'source> ParsingContext<'source> {
         //    type_qualifier SEMICOLON type_qualifier IDENTIFIER SEMICOLON
         //    type_qualifier IDENTIFIER identifier_list SEMICOLON
 
-        if self.peek_type_qualifier(parser) || self.peek_type_name(parser) {
-            let qualifiers = self.parse_type_qualifiers(parser)?;
+        if self.peek_type_qualifier(frontend) || self.peek_type_name(frontend) {
+            let mut qualifiers = self.parse_type_qualifiers(frontend, ctx)?;
 
-            if self.peek_type_name(parser) {
+            if self.peek_type_name(frontend) {
                 // This branch handles variables and function prototypes and if
                 // external is true also function definitions
-                let (ty, mut meta) = self.parse_type(parser)?;
+                let (ty, mut meta) = self.parse_type(frontend, ctx)?;
 
-                let token = self.bump(parser)?;
+                let token = self.bump(frontend)?;
                 let token_fallthrough = match token.value {
-                    TokenValue::Identifier(name) => match self.expect_peek(parser)?.value {
+                    TokenValue::Identifier(name) => match self.expect_peek(frontend)?.value {
                         TokenValue::LeftParen => {
                             // This branch handles function definition and prototypes
-                            self.bump(parser)?;
+                            self.bump(frontend)?;
 
                             let result = ty.map(|ty| FunctionResult { ty, binding: None });
-                            let mut body = Block::new();
 
-                            let mut context = Context::new(parser, &mut body);
+                            let mut context = Context::new(frontend, ctx.module, false)?;
 
-                            self.parse_function_args(parser, &mut context, &mut body)?;
+                            self.parse_function_args(frontend, &mut context)?;
 
-                            let end_meta = self.expect(parser, TokenValue::RightParen)?.meta;
+                            let end_meta = self.expect(frontend, TokenValue::RightParen)?.meta;
                             meta.subsume(end_meta);
 
-                            let token = self.bump(parser)?;
+                            let token = self.bump(frontend)?;
                             return match token.value {
                                 TokenValue::Semicolon => {
                                     // This branch handles function prototypes
-                                    parser.add_prototype(context, name, result, meta);
+                                    frontend.add_prototype(context, name, result, meta);
 
                                     Ok(Some(meta))
                                 }
@@ -289,13 +354,13 @@ impl<'source> ParsingContext<'source> {
                                     // parse the body
                                     self.parse_compound_statement(
                                         token.meta,
-                                        parser,
+                                        frontend,
                                         &mut context,
-                                        &mut body,
                                         &mut None,
+                                        false,
                                     )?;
 
-                                    parser.add_function(context, name, result, body, meta);
+                                    frontend.add_function(context, name, result, meta);
 
                                     Ok(Some(meta))
                                 }
@@ -318,30 +383,31 @@ impl<'source> ParsingContext<'source> {
                                 }),
                             };
                         }
-                        // Pass the token to the init_declator_list parser
+                        // Pass the token to the init_declarator_list parser
                         _ => Token {
                             value: TokenValue::Identifier(name),
                             meta: token.meta,
                         },
                     },
-                    // Pass the token to the init_declator_list parser
+                    // Pass the token to the init_declarator_list parser
                     _ => token,
                 };
 
                 // If program execution has reached here then this will be a
                 // init_declarator_list
-                // token_falltrough will have a token that was already bumped
+                // token_fallthrough will have a token that was already bumped
                 if let Some(ty) = ty {
                     let mut ctx = DeclarationContext {
                         qualifiers,
                         external,
+                        is_inside_loop,
                         ctx,
-                        body,
                     };
 
-                    self.parse_init_declarator_list(parser, ty, Some(token_fallthrough), &mut ctx)?;
+                    self.backtrack(token_fallthrough)?;
+                    self.parse_init_declarator_list(frontend, ty, &mut ctx)?;
                 } else {
-                    parser.errors.push(Error {
+                    frontend.errors.push(Error {
                         kind: ErrorKind::SemanticError("Declaration cannot have void type".into()),
                         meta,
                     })
@@ -353,20 +419,27 @@ impl<'source> ParsingContext<'source> {
                 // ```glsl
                 // layout(early_fragment_tests);
                 // ```
-                let token = self.bump(parser)?;
+                let token = self.bump(frontend)?;
                 match token.value {
                     TokenValue::Identifier(ty_name) => {
-                        if self.bump_if(parser, TokenValue::LeftBrace).is_some() {
+                        if self.bump_if(frontend, TokenValue::LeftBrace).is_some() {
                             self.parse_block_declaration(
-                                parser,
+                                frontend,
                                 ctx,
-                                body,
-                                &qualifiers,
+                                &mut qualifiers,
                                 ty_name,
                                 token.meta,
                             )
                             .map(Some)
                         } else {
+                            if qualifiers.invariant.take().is_some() {
+                                frontend.make_variable_invariant(ctx, &ty_name, token.meta)?;
+
+                                qualifiers.unused_errors(&mut frontend.errors);
+                                self.expect(frontend, TokenValue::Semicolon)?;
+                                return Ok(Some(qualifiers.span));
+                            }
+
                             //TODO: declaration
                             // type_qualifier IDENTIFIER SEMICOLON
                             // type_qualifier IDENTIFIER identifier_list SEMICOLON
@@ -377,33 +450,28 @@ impl<'source> ParsingContext<'source> {
                         }
                     }
                     TokenValue::Semicolon => {
-                        let mut meta_all = token.meta;
-                        for &(ref qualifier, meta) in qualifiers.iter() {
-                            meta_all.subsume(meta);
-                            match *qualifier {
-                                TypeQualifier::WorkGroupSize(i, value) => {
-                                    parser.meta.workgroup_size[i] = value
-                                }
-                                TypeQualifier::EarlyFragmentTests => {
-                                    parser.meta.early_fragment_tests = true;
-                                }
-                                TypeQualifier::StorageQualifier(_) => {
-                                    // TODO: Maybe add some checks here
-                                    // This is needed because of cases like
-                                    // layout(early_fragment_tests) in;
-                                }
-                                _ => {
-                                    parser.errors.push(Error {
-                                        kind: ErrorKind::SemanticError(
-                                            "Qualifier not supported as standalone".into(),
-                                        ),
-                                        meta,
-                                    });
-                                }
-                            }
+                        if let Some(value) =
+                            qualifiers.uint_layout_qualifier("local_size_x", &mut frontend.errors)
+                        {
+                            frontend.meta.workgroup_size[0] = value;
+                        }
+                        if let Some(value) =
+                            qualifiers.uint_layout_qualifier("local_size_y", &mut frontend.errors)
+                        {
+                            frontend.meta.workgroup_size[1] = value;
+                        }
+                        if let Some(value) =
+                            qualifiers.uint_layout_qualifier("local_size_z", &mut frontend.errors)
+                        {
+                            frontend.meta.workgroup_size[2] = value;
                         }
 
-                        Ok(Some(meta_all))
+                        frontend.meta.early_fragment_tests |= qualifiers
+                            .none_layout_qualifier("early_fragment_tests", &mut frontend.errors);
+
+                        qualifiers.unused_errors(&mut frontend.errors);
+
+                        Ok(Some(qualifiers.span))
                     }
                     _ => Err(Error {
                         kind: ErrorKind::InvalidToken(
@@ -415,12 +483,12 @@ impl<'source> ParsingContext<'source> {
                 }
             }
         } else {
-            match self.peek(parser).map(|t| &t.value) {
+            match self.peek(frontend).map(|t| &t.value) {
                 Some(&TokenValue::Precision) => {
                     // PRECISION precision_qualifier type_specifier SEMICOLON
-                    self.bump(parser)?;
+                    self.bump(frontend)?;
 
-                    let token = self.bump(parser)?;
+                    let token = self.bump(frontend)?;
                     let _ = match token.value {
                         TokenValue::PrecisionQualifier(p) => p,
                         _ => {
@@ -438,18 +506,14 @@ impl<'source> ParsingContext<'source> {
                         }
                     };
 
-                    let (ty, meta) = self.parse_type_non_void(parser)?;
+                    let (ty, meta) = self.parse_type_non_void(frontend, ctx)?;
 
-                    match parser.module.types[ty].inner {
+                    match ctx.module.types[ty].inner {
                         TypeInner::Scalar {
-                            kind: ScalarKind::Float,
-                            ..
-                        }
-                        | TypeInner::Scalar {
-                            kind: ScalarKind::Sint,
+                            kind: ScalarKind::Float | ScalarKind::Sint,
                             ..
                         } => {}
-                        _ => parser.errors.push(Error {
+                        _ => frontend.errors.push(Error {
                             kind: ErrorKind::SemanticError(
                                 "Precision statement can only work on floats and ints".into(),
                             ),
@@ -457,7 +521,7 @@ impl<'source> ParsingContext<'source> {
                         }),
                     }
 
-                    self.expect(parser, TokenValue::Semicolon)?;
+                    self.expect(frontend, TokenValue::Semicolon)?;
 
                     Ok(Some(meta))
                 }
@@ -468,37 +532,31 @@ impl<'source> ParsingContext<'source> {
 
     pub fn parse_block_declaration(
         &mut self,
-        parser: &mut Parser,
+        frontend: &mut Frontend,
         ctx: &mut Context,
-        body: &mut Block,
-        qualifiers: &[(TypeQualifier, Span)],
+        qualifiers: &mut TypeQualifiers,
         ty_name: String,
-        meta: Span,
+        mut meta: Span,
     ) -> Result<Span> {
-        let mut storage = None;
-        let mut layout = None;
-
-        for &(ref qualifier, _) in qualifiers {
-            match *qualifier {
-                TypeQualifier::StorageQualifier(StorageQualifier::StorageClass(c)) => {
-                    storage = Some(c)
+        let layout = match qualifiers.layout_qualifiers.remove(&QualifierKey::Layout) {
+            Some((QualifierValue::Layout(l), _)) => l,
+            None => {
+                if let StorageQualifier::AddressSpace(AddressSpace::Storage { .. }) =
+                    qualifiers.storage.0
+                {
+                    StructLayout::Std430
+                } else {
+                    StructLayout::Std140
                 }
-                TypeQualifier::Layout(l) => layout = Some(l),
-                _ => continue,
             }
-        }
-
-        let layout = match (layout, storage) {
-            (Some(layout), _) => layout,
-            (None, Some(StorageClass::Storage { .. })) => StructLayout::Std430,
-            _ => StructLayout::Std140,
+            _ => unreachable!(),
         };
 
         let mut members = Vec::new();
-        let span = self.parse_struct_declaration_list(parser, &mut members, layout)?;
-        self.expect(parser, TokenValue::RightBrace)?;
+        let span = self.parse_struct_declaration_list(frontend, ctx, &mut members, layout)?;
+        self.expect(frontend, TokenValue::RightBrace)?;
 
-        let mut ty = parser.module.types.insert(
+        let mut ty = ctx.module.types.insert(
             Type {
                 name: Some(ty_name),
                 inner: TypeInner::Struct {
@@ -509,14 +567,13 @@ impl<'source> ParsingContext<'source> {
             Default::default(),
         );
 
-        let token = self.bump(parser)?;
+        let token = self.bump(frontend)?;
         let name = match token.value {
             TokenValue::Semicolon => None,
             TokenValue::Identifier(name) => {
-                let array_specifier = self.parse_array_specifier(parser)?;
-                ty = parser.maybe_array(ty, token.meta, array_specifier);
+                self.parse_array_specifier(frontend, ctx, &mut meta, &mut ty)?;
 
-                self.expect(parser, TokenValue::Semicolon)?;
+                self.expect(frontend, TokenValue::Semicolon)?;
 
                 Some(name)
             }
@@ -531,9 +588,8 @@ impl<'source> ParsingContext<'source> {
             }
         };
 
-        let global = parser.add_global_var(
+        let global = frontend.add_global_var(
             ctx,
-            body,
             VarDeclaration {
                 qualifiers,
                 ty,
@@ -555,9 +611,9 @@ impl<'source> ParsingContext<'source> {
                 entry_arg: None,
                 mutable: true,
             };
-            ctx.add_global(parser, &k, lookup, body);
+            ctx.add_global(&k, lookup)?;
 
-            parser.global_variables.push((k, lookup));
+            frontend.global_variables.push((k, lookup));
         }
 
         Ok(meta)
@@ -566,53 +622,60 @@ impl<'source> ParsingContext<'source> {
     // TODO: Accept layout arguments
     pub fn parse_struct_declaration_list(
         &mut self,
-        parser: &mut Parser,
+        frontend: &mut Frontend,
+        ctx: &mut Context,
         members: &mut Vec<StructMember>,
         layout: StructLayout,
     ) -> Result<u32> {
         let mut span = 0;
-        let mut align = 0;
+        let mut align = Alignment::ONE;
 
         loop {
             // TODO: type_qualifier
 
-            let (ty, mut meta) = self.parse_type_non_void(parser)?;
-            let (name, end_meta) = self.expect_ident(parser)?;
+            let (base_ty, mut meta) = self.parse_type_non_void(frontend, ctx)?;
 
-            meta.subsume(end_meta);
+            loop {
+                let (name, name_meta) = self.expect_ident(frontend)?;
+                let mut ty = base_ty;
+                self.parse_array_specifier(frontend, ctx, &mut meta, &mut ty)?;
 
-            let array_specifier = self.parse_array_specifier(parser)?;
-            let ty = parser.maybe_array(ty, meta, array_specifier);
+                meta.subsume(name_meta);
 
-            self.expect(parser, TokenValue::Semicolon)?;
+                let info = offset::calculate_offset(
+                    ty,
+                    meta,
+                    layout,
+                    &mut ctx.module.types,
+                    &mut frontend.errors,
+                );
 
-            let info = offset::calculate_offset(
-                ty,
-                meta,
-                layout,
-                &mut parser.module.types,
-                &parser.module.constants,
-                &mut parser.errors,
-            );
+                let member_alignment = info.align;
+                span = member_alignment.round_up(span);
+                align = member_alignment.max(align);
 
-            span = crate::front::align_up(span, info.align);
-            align = align.max(info.align);
+                members.push(StructMember {
+                    name: Some(name),
+                    ty: info.ty,
+                    binding: None,
+                    offset: span,
+                });
 
-            members.push(StructMember {
-                name: Some(name),
-                ty: info.ty,
-                binding: None,
-                offset: span,
-            });
+                span += info.span;
 
-            span += info.span;
+                if self.bump_if(frontend, TokenValue::Comma).is_none() {
+                    break;
+                }
+            }
 
-            if let TokenValue::RightBrace = self.expect_peek(parser)?.value {
+            self.expect(frontend, TokenValue::Semicolon)?;
+
+            if let TokenValue::RightBrace = self.expect_peek(frontend)?.value {
                 break;
             }
         }
 
-        span = crate::front::align_up(span, align);
+        span = align.round_up(span);
 
         Ok(span)
     }
