@@ -54,7 +54,7 @@ use thiserror::Error;
 /// })
 /// ```
 ///
-/// and the type of an access to a pointer of storage class `class` to such a
+/// and the type of an access to a pointer of address space `space` to such a
 /// matrix might be:
 ///
 /// ```ignore
@@ -62,7 +62,7 @@ use thiserror::Error;
 ///     size: Some(rows),
 ///     kind: ScalarKind::Float,
 ///     width,
-///     class
+///     space,
 /// })
 /// ```
 ///
@@ -97,7 +97,7 @@ pub enum TypeResolution {
 }
 
 impl TypeResolution {
-    pub fn handle(&self) -> Option<Handle<crate::Type>> {
+    pub const fn handle(&self) -> Option<Handle<crate::Type>> {
         match *self {
             Self::Handle(handle) => Some(handle),
             Self::Value(_) => None,
@@ -130,17 +130,17 @@ impl Clone for TypeResolution {
                     columns,
                     width,
                 },
-                Ti::Pointer { base, class } => Ti::Pointer { base, class },
+                Ti::Pointer { base, space } => Ti::Pointer { base, space },
                 Ti::ValuePointer {
                     size,
                     kind,
                     width,
-                    class,
+                    space,
                 } => Ti::ValuePointer {
                     size,
                     kind,
                     width,
-                    class,
+                    space,
                 },
                 _ => unreachable!("Unexpected clone type: {:?}", v),
             }),
@@ -148,18 +148,6 @@ impl Clone for TypeResolution {
     }
 }
 
-impl crate::ConstantInner {
-    pub fn resolve_type(&self) -> TypeResolution {
-        match *self {
-            Self::Scalar { width, ref value } => TypeResolution::Value(crate::TypeInner::Scalar {
-                kind: value.scalar_kind(),
-                width,
-            }),
-            Self::Composite { ty, components: _ } => TypeResolution::Handle(ty),
-        }
-    }
-}
-
 #[derive(Clone, Debug, Error, PartialEq)]
 pub enum ResolveError {
     #[error("Index {index} is out of bounds for expression {expr:?}")]
@@ -189,15 +177,18 @@ pub enum ResolveError {
     FunctionNotDefined { name: String },
     #[error("Function without return type")]
     FunctionReturnsVoid,
-    #[error("Type is not found in the given immutable arena")]
-    TypeNotFound,
     #[error("Incompatible operands: {0}")]
     IncompatibleOperands(String),
+    #[error("Function argument {0} doesn't exist")]
+    FunctionArgumentNotFound(u32),
+    #[error("Special type is not registered within the module")]
+    MissingSpecialType,
 }
 
 pub struct ResolveContext<'a> {
     pub constants: &'a Arena<crate::Constant>,
     pub types: &'a UniqueArena<crate::Type>,
+    pub special_types: &'a crate::SpecialTypes,
     pub global_vars: &'a Arena<crate::GlobalVariable>,
     pub local_vars: &'a Arena<crate::LocalVariable>,
     pub functions: &'a Arena<crate::Function>,
@@ -205,6 +196,23 @@ pub struct ResolveContext<'a> {
 }
 
 impl<'a> ResolveContext<'a> {
+    /// Initialize a resolve context from the module.
+    pub const fn with_locals(
+        module: &'a crate::Module,
+        local_vars: &'a Arena<crate::LocalVariable>,
+        arguments: &'a [crate::FunctionArgument],
+    ) -> Self {
+        Self {
+            constants: &module.constants,
+            types: &module.types,
+            special_types: &module.special_types,
+            global_vars: &module.global_variables,
+            local_vars,
+            functions: &module.functions,
+            arguments,
+        }
+    }
+
     /// Determine the type of `expr`.
     ///
     /// The `past` argument must be a closure that can resolve the types of any
@@ -223,12 +231,12 @@ impl<'a> ResolveContext<'a> {
     pub fn resolve(
         &self,
         expr: &crate::Expression,
-        past: impl Fn(Handle<crate::Expression>) -> &'a TypeResolution,
+        past: impl Fn(Handle<crate::Expression>) -> Result<&'a TypeResolution, ResolveError>,
     ) -> Result<TypeResolution, ResolveError> {
         use crate::TypeInner as Ti;
         let types = self.types;
         Ok(match *expr {
-            crate::Expression::Access { base, .. } => match *past(base).inner_with(types) {
+            crate::Expression::Access { base, .. } => match *past(base)?.inner_with(types) {
                 // Arrays and matrices can only be indexed dynamically behind a
                 // pointer, but that's a validation error, not a type error, so
                 // go ahead provide a type here.
@@ -247,16 +255,16 @@ impl<'a> ResolveContext<'a> {
                     size: Some(_),
                     kind,
                     width,
-                    class,
+                    space,
                 } => TypeResolution::Value(Ti::ValuePointer {
                     size: None,
                     kind,
                     width,
-                    class,
+                    space,
                 }),
-                Ti::Pointer { base, class } => {
+                Ti::Pointer { base, space } => {
                     TypeResolution::Value(match types[base].inner {
-                        Ti::Array { base, .. } => Ti::Pointer { base, class },
+                        Ti::Array { base, .. } => Ti::Pointer { base, space },
                         Ti::Vector {
                             size: _,
                             kind,
@@ -265,7 +273,7 @@ impl<'a> ResolveContext<'a> {
                             size: None,
                             kind,
                             width,
-                            class,
+                            space,
                         },
                         // Matrices are only dynamically indexed behind a pointer
                         Ti::Matrix {
@@ -276,8 +284,9 @@ impl<'a> ResolveContext<'a> {
                             kind: crate::ScalarKind::Float,
                             size: Some(rows),
                             width,
-                            class,
+                            space,
                         },
+                        Ti::BindingArray { base, .. } => Ti::Pointer { base, space },
                         ref other => {
                             log::error!("Access sub-type {:?}", other);
                             return Err(ResolveError::InvalidSubAccess {
@@ -287,6 +296,7 @@ impl<'a> ResolveContext<'a> {
                         }
                     })
                 }
+                Ti::BindingArray { base, .. } => TypeResolution::Handle(base),
                 ref other => {
                     log::error!("Access type {:?}", other);
                     return Err(ResolveError::InvalidAccess {
@@ -295,116 +305,111 @@ impl<'a> ResolveContext<'a> {
                     });
                 }
             },
-            crate::Expression::AccessIndex { base, index } => match *past(base).inner_with(types) {
-                Ti::Vector { size, kind, width } => {
-                    if index >= size as u32 {
-                        return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
-                    }
-                    TypeResolution::Value(Ti::Scalar { kind, width })
-                }
-                Ti::Matrix {
-                    columns,
-                    rows,
-                    width,
-                } => {
-                    if index >= columns as u32 {
-                        return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
-                    }
-                    TypeResolution::Value(crate::TypeInner::Vector {
-                        size: rows,
-                        kind: crate::ScalarKind::Float,
-                        width,
-                    })
-                }
-                Ti::Array { base, .. } => TypeResolution::Handle(base),
-                Ti::Struct { ref members, .. } => {
-                    let member = members
-                        .get(index as usize)
-                        .ok_or(ResolveError::OutOfBoundsIndex { expr: base, index })?;
-                    TypeResolution::Handle(member.ty)
-                }
-                Ti::ValuePointer {
-                    size: Some(size),
-                    kind,
-                    width,
-                    class,
-                } => {
-                    if index >= size as u32 {
-                        return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
-                    }
-                    TypeResolution::Value(Ti::ValuePointer {
-                        size: None,
-                        kind,
-                        width,
-                        class,
-                    })
-                }
-                Ti::Pointer {
-                    base: ty_base,
-                    class,
-                } => TypeResolution::Value(match types[ty_base].inner {
-                    Ti::Array { base, .. } => Ti::Pointer { base, class },
+            crate::Expression::AccessIndex { base, index } => {
+                match *past(base)?.inner_with(types) {
                     Ti::Vector { size, kind, width } => {
                         if index >= size as u32 {
                             return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
                         }
-                        Ti::ValuePointer {
-                            size: None,
-                            kind,
-                            width,
-                            class,
-                        }
+                        TypeResolution::Value(Ti::Scalar { kind, width })
                     }
                     Ti::Matrix {
-                        rows,
                         columns,
+                        rows,
                         width,
                     } => {
                         if index >= columns as u32 {
                             return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
                         }
-                        Ti::ValuePointer {
-                            size: Some(rows),
+                        TypeResolution::Value(crate::TypeInner::Vector {
+                            size: rows,
                             kind: crate::ScalarKind::Float,
                             width,
-                            class,
-                        }
+                        })
                     }
+                    Ti::Array { base, .. } => TypeResolution::Handle(base),
                     Ti::Struct { ref members, .. } => {
                         let member = members
                             .get(index as usize)
                             .ok_or(ResolveError::OutOfBoundsIndex { expr: base, index })?;
-                        Ti::Pointer {
-                            base: member.ty,
-                            class,
-                        }
+                        TypeResolution::Handle(member.ty)
                     }
+                    Ti::ValuePointer {
+                        size: Some(size),
+                        kind,
+                        width,
+                        space,
+                    } => {
+                        if index >= size as u32 {
+                            return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
+                        }
+                        TypeResolution::Value(Ti::ValuePointer {
+                            size: None,
+                            kind,
+                            width,
+                            space,
+                        })
+                    }
+                    Ti::Pointer {
+                        base: ty_base,
+                        space,
+                    } => TypeResolution::Value(match types[ty_base].inner {
+                        Ti::Array { base, .. } => Ti::Pointer { base, space },
+                        Ti::Vector { size, kind, width } => {
+                            if index >= size as u32 {
+                                return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
+                            }
+                            Ti::ValuePointer {
+                                size: None,
+                                kind,
+                                width,
+                                space,
+                            }
+                        }
+                        Ti::Matrix {
+                            rows,
+                            columns,
+                            width,
+                        } => {
+                            if index >= columns as u32 {
+                                return Err(ResolveError::OutOfBoundsIndex { expr: base, index });
+                            }
+                            Ti::ValuePointer {
+                                size: Some(rows),
+                                kind: crate::ScalarKind::Float,
+                                width,
+                                space,
+                            }
+                        }
+                        Ti::Struct { ref members, .. } => {
+                            let member = members
+                                .get(index as usize)
+                                .ok_or(ResolveError::OutOfBoundsIndex { expr: base, index })?;
+                            Ti::Pointer {
+                                base: member.ty,
+                                space,
+                            }
+                        }
+                        Ti::BindingArray { base, .. } => Ti::Pointer { base, space },
+                        ref other => {
+                            log::error!("Access index sub-type {:?}", other);
+                            return Err(ResolveError::InvalidSubAccess {
+                                ty: ty_base,
+                                indexed: true,
+                            });
+                        }
+                    }),
+                    Ti::BindingArray { base, .. } => TypeResolution::Handle(base),
                     ref other => {
-                        log::error!("Access index sub-type {:?}", other);
-                        return Err(ResolveError::InvalidSubAccess {
-                            ty: ty_base,
+                        log::error!("Access index type {:?}", other);
+                        return Err(ResolveError::InvalidAccess {
+                            expr: base,
                             indexed: true,
                         });
                     }
-                }),
-                ref other => {
-                    log::error!("Access index type {:?}", other);
-                    return Err(ResolveError::InvalidAccess {
-                        expr: base,
-                        indexed: true,
-                    });
                 }
-            },
-            crate::Expression::Constant(h) => match self.constants[h].inner {
-                crate::ConstantInner::Scalar { width, ref value } => {
-                    TypeResolution::Value(Ti::Scalar {
-                        kind: value.scalar_kind(),
-                        width,
-                    })
-                }
-                crate::ConstantInner::Composite { ty, components: _ } => TypeResolution::Handle(ty),
-            },
-            crate::Expression::Splat { size, value } => match *past(value).inner_with(types) {
+            }
+            crate::Expression::Splat { size, value } => match *past(value)?.inner_with(types) {
                 Ti::Scalar { kind, width } => {
                     TypeResolution::Value(Ti::Vector { size, kind, width })
                 }
@@ -417,7 +422,7 @@ impl<'a> ResolveContext<'a> {
                 size,
                 vector,
                 pattern: _,
-            } => match *past(vector).inner_with(types) {
+            } => match *past(vector)?.inner_with(types) {
                 Ti::Vector {
                     size: _,
                     kind,
@@ -428,18 +433,25 @@ impl<'a> ResolveContext<'a> {
                     return Err(ResolveError::InvalidVector(vector));
                 }
             },
+            crate::Expression::Literal(lit) => TypeResolution::Value(lit.ty_inner()),
+            crate::Expression::Constant(h) => TypeResolution::Handle(self.constants[h].ty),
+            crate::Expression::ZeroValue(ty) => TypeResolution::Handle(ty),
             crate::Expression::Compose { ty, .. } => TypeResolution::Handle(ty),
             crate::Expression::FunctionArgument(index) => {
-                TypeResolution::Handle(self.arguments[index as usize].ty)
+                let arg = self
+                    .arguments
+                    .get(index as usize)
+                    .ok_or(ResolveError::FunctionArgumentNotFound(index))?;
+                TypeResolution::Handle(arg.ty)
             }
             crate::Expression::GlobalVariable(h) => {
                 let var = &self.global_vars[h];
-                if var.class == crate::StorageClass::Handle {
+                if var.space == crate::AddressSpace::Handle {
                     TypeResolution::Handle(var.ty)
                 } else {
                     TypeResolution::Value(Ti::Pointer {
                         base: var.ty,
-                        class: var.class,
+                        space: var.space,
                     })
                 }
             }
@@ -447,11 +459,11 @@ impl<'a> ResolveContext<'a> {
                 let var = &self.local_vars[h];
                 TypeResolution::Value(Ti::Pointer {
                     base: var.ty,
-                    class: crate::StorageClass::Function,
+                    space: crate::AddressSpace::Function,
                 })
             }
-            crate::Expression::Load { pointer } => match *past(pointer).inner_with(types) {
-                Ti::Pointer { base, class: _ } => {
+            crate::Expression::Load { pointer } => match *past(pointer)?.inner_with(types) {
+                Ti::Pointer { base, space: _ } => {
                     if let Ti::Atomic { kind, width } = types[base].inner {
                         TypeResolution::Value(Ti::Scalar { kind, width })
                     } else {
@@ -462,7 +474,7 @@ impl<'a> ResolveContext<'a> {
                     size,
                     kind,
                     width,
-                    class: _,
+                    space: _,
                 } => TypeResolution::Value(match size {
                     Some(size) => Ti::Vector { size, kind, width },
                     None => Ti::Scalar { kind, width },
@@ -476,7 +488,7 @@ impl<'a> ResolveContext<'a> {
                 image,
                 gather: Some(_),
                 ..
-            } => match *past(image).inner_with(types) {
+            } => match *past(image)?.inner_with(types) {
                 Ti::Image { class, .. } => TypeResolution::Value(Ti::Vector {
                     kind: match class {
                         crate::ImageClass::Sampled { kind, multi: _ } => kind,
@@ -491,7 +503,7 @@ impl<'a> ResolveContext<'a> {
                 }
             },
             crate::Expression::ImageSample { image, .. }
-            | crate::Expression::ImageLoad { image, .. } => match *past(image).inner_with(types) {
+            | crate::Expression::ImageLoad { image, .. } => match *past(image)?.inner_with(types) {
                 Ti::Image { class, .. } => TypeResolution::Value(match class {
                     crate::ImageClass::Depth { multi: _ } => Ti::Scalar {
                         kind: crate::ScalarKind::Float,
@@ -514,20 +526,20 @@ impl<'a> ResolveContext<'a> {
                 }
             },
             crate::Expression::ImageQuery { image, query } => TypeResolution::Value(match query {
-                crate::ImageQuery::Size { level: _ } => match *past(image).inner_with(types) {
+                crate::ImageQuery::Size { level: _ } => match *past(image)?.inner_with(types) {
                     Ti::Image { dim, .. } => match dim {
                         crate::ImageDimension::D1 => Ti::Scalar {
-                            kind: crate::ScalarKind::Sint,
+                            kind: crate::ScalarKind::Uint,
                             width: 4,
                         },
                         crate::ImageDimension::D2 | crate::ImageDimension::Cube => Ti::Vector {
                             size: crate::VectorSize::Bi,
-                            kind: crate::ScalarKind::Sint,
+                            kind: crate::ScalarKind::Uint,
                             width: 4,
                         },
                         crate::ImageDimension::D3 => Ti::Vector {
                             size: crate::VectorSize::Tri,
-                            kind: crate::ScalarKind::Sint,
+                            kind: crate::ScalarKind::Uint,
                             width: 4,
                         },
                     },
@@ -539,18 +551,18 @@ impl<'a> ResolveContext<'a> {
                 crate::ImageQuery::NumLevels
                 | crate::ImageQuery::NumLayers
                 | crate::ImageQuery::NumSamples => Ti::Scalar {
-                    kind: crate::ScalarKind::Sint,
+                    kind: crate::ScalarKind::Uint,
                     width: 4,
                 },
             }),
-            crate::Expression::Unary { expr, .. } => past(expr).clone(),
+            crate::Expression::Unary { expr, .. } => past(expr)?.clone(),
             crate::Expression::Binary { op, left, right } => match op {
                 crate::BinaryOperator::Add
                 | crate::BinaryOperator::Subtract
                 | crate::BinaryOperator::Divide
-                | crate::BinaryOperator::Modulo => past(left).clone(),
+                | crate::BinaryOperator::Modulo => past(left)?.clone(),
                 crate::BinaryOperator::Multiply => {
-                    let (res_left, res_right) = (past(left), past(right));
+                    let (res_left, res_right) = (past(left)?, past(right)?);
                     match (res_left.inner_with(types), res_right.inner_with(types)) {
                         (
                             &Ti::Matrix {
@@ -593,8 +605,7 @@ impl<'a> ResolveContext<'a> {
                         (&Ti::Vector { .. }, &Ti::Vector { .. }) => res_left.clone(),
                         (tl, tr) => {
                             return Err(ResolveError::IncompatibleOperands(format!(
-                                "{:?} * {:?}",
-                                tl, tr
+                                "{tl:?} * {tr:?}"
                             )))
                         }
                     }
@@ -609,13 +620,12 @@ impl<'a> ResolveContext<'a> {
                 | crate::BinaryOperator::LogicalOr => {
                     let kind = crate::ScalarKind::Bool;
                     let width = crate::BOOL_WIDTH;
-                    let inner = match *past(left).inner_with(types) {
+                    let inner = match *past(left)?.inner_with(types) {
                         Ti::Scalar { .. } => Ti::Scalar { kind, width },
                         Ti::Vector { size, .. } => Ti::Vector { size, kind, width },
                         ref other => {
                             return Err(ResolveError::IncompatibleOperands(format!(
-                                "{:?}({:?}, _)",
-                                op, other
+                                "{op:?}({other:?}, _)"
                             )))
                         }
                     };
@@ -625,25 +635,12 @@ impl<'a> ResolveContext<'a> {
                 | crate::BinaryOperator::ExclusiveOr
                 | crate::BinaryOperator::InclusiveOr
                 | crate::BinaryOperator::ShiftLeft
-                | crate::BinaryOperator::ShiftRight => past(left).clone(),
+                | crate::BinaryOperator::ShiftRight => past(left)?.clone(),
             },
-            crate::Expression::AtomicResult {
-                kind,
-                width,
-                comparison,
-            } => {
-                if comparison {
-                    TypeResolution::Value(Ti::Vector {
-                        size: crate::VectorSize::Bi,
-                        kind,
-                        width,
-                    })
-                } else {
-                    TypeResolution::Value(Ti::Scalar { kind, width })
-                }
-            }
-            crate::Expression::Select { accept, .. } => past(accept).clone(),
-            crate::Expression::Derivative { axis: _, expr } => past(expr).clone(),
+            crate::Expression::AtomicResult { ty, .. } => TypeResolution::Handle(ty),
+            crate::Expression::WorkGroupUniformLoadResult { ty } => TypeResolution::Handle(ty),
+            crate::Expression::Select { accept, .. } => past(accept)?.clone(),
+            crate::Expression::Derivative { expr, .. } => past(expr)?.clone(),
             crate::Expression::Relational { fun, argument } => match fun {
                 crate::RelationalFunction::All | crate::RelationalFunction::Any => {
                     TypeResolution::Value(Ti::Scalar {
@@ -651,26 +648,24 @@ impl<'a> ResolveContext<'a> {
                         width: crate::BOOL_WIDTH,
                     })
                 }
-                crate::RelationalFunction::IsNan
-                | crate::RelationalFunction::IsInf
-                | crate::RelationalFunction::IsFinite
-                | crate::RelationalFunction::IsNormal => match *past(argument).inner_with(types) {
-                    Ti::Scalar { .. } => TypeResolution::Value(Ti::Scalar {
-                        kind: crate::ScalarKind::Bool,
-                        width: crate::BOOL_WIDTH,
-                    }),
-                    Ti::Vector { size, .. } => TypeResolution::Value(Ti::Vector {
-                        kind: crate::ScalarKind::Bool,
-                        width: crate::BOOL_WIDTH,
-                        size,
-                    }),
-                    ref other => {
-                        return Err(ResolveError::IncompatibleOperands(format!(
-                            "{:?}({:?})",
-                            fun, other
-                        )))
+                crate::RelationalFunction::IsNan | crate::RelationalFunction::IsInf => {
+                    match *past(argument)?.inner_with(types) {
+                        Ti::Scalar { .. } => TypeResolution::Value(Ti::Scalar {
+                            kind: crate::ScalarKind::Bool,
+                            width: crate::BOOL_WIDTH,
+                        }),
+                        Ti::Vector { size, .. } => TypeResolution::Value(Ti::Vector {
+                            kind: crate::ScalarKind::Bool,
+                            width: crate::BOOL_WIDTH,
+                            size,
+                        }),
+                        ref other => {
+                            return Err(ResolveError::IncompatibleOperands(format!(
+                                "{fun:?}({other:?})"
+                            )))
+                        }
                     }
-                },
+                }
             },
             crate::Expression::Math {
                 fun,
@@ -680,13 +675,14 @@ impl<'a> ResolveContext<'a> {
                 arg3: _,
             } => {
                 use crate::MathFunction as Mf;
-                let res_arg = past(arg);
+                let res_arg = past(arg)?;
                 match fun {
                     // comparison
                     Mf::Abs |
                     Mf::Min |
                     Mf::Max |
                     Mf::Clamp |
+                    Mf::Saturate |
                     // trigonometry
                     Mf::Cos |
                     Mf::Cosh |
@@ -701,14 +697,14 @@ impl<'a> ResolveContext<'a> {
                     Mf::Asinh |
                     Mf::Acosh |
                     Mf::Atanh |
+                    Mf::Radians |
+                    Mf::Degrees |
                     // decomposition
                     Mf::Ceil |
                     Mf::Floor |
                     Mf::Round |
                     Mf::Fract |
                     Mf::Trunc |
-                    Mf::Modf |
-                    Mf::Frexp |
                     Mf::Ldexp |
                     // exponent
                     Mf::Exp |
@@ -716,6 +712,31 @@ impl<'a> ResolveContext<'a> {
                     Mf::Log |
                     Mf::Log2 |
                     Mf::Pow => res_arg.clone(),
+                    Mf::Modf | Mf::Frexp => {
+                        let (size, width) = match res_arg.inner_with(types) {
+                            &Ti::Scalar {
+                                kind: crate::ScalarKind::Float,
+                               width,
+                            } => (None, width),
+                            &Ti::Vector {
+                                kind: crate::ScalarKind::Float,
+                                size,
+                                width,
+                            } => (Some(size), width),
+                            ref other =>
+                                return Err(ResolveError::IncompatibleOperands(format!("{fun:?}({other:?}, _)")))
+                        };
+                        let result = self
+                        .special_types
+                        .predeclared_types
+                        .get(&if fun == Mf::Modf {
+                            crate::PredeclaredType::ModfResult { size, width }
+                    } else {
+                            crate::PredeclaredType::FrexpResult { size, width }
+                    })
+                        .ok_or(ResolveError::MissingSpecialType)?;
+                        TypeResolution::Handle(*result)
+                    },
                     // geometry
                     Mf::Dot => match *res_arg.inner_with(types) {
                         Ti::Vector {
@@ -725,18 +746,18 @@ impl<'a> ResolveContext<'a> {
                         } => TypeResolution::Value(Ti::Scalar { kind, width }),
                         ref other =>
                             return Err(ResolveError::IncompatibleOperands(
-                                format!("{:?}({:?}, _)", fun, other)
+                                format!("{fun:?}({other:?}, _)")
                             )),
                     },
                     Mf::Outer => {
                         let arg1 = arg1.ok_or_else(|| ResolveError::IncompatibleOperands(
-                            format!("{:?}(_, None)", fun)
+                            format!("{fun:?}(_, None)")
                         ))?;
-                        match (res_arg.inner_with(types), past(arg1).inner_with(types)) {
+                        match (res_arg.inner_with(types), past(arg1)?.inner_with(types)) {
                             (&Ti::Vector {kind: _, size: columns,width}, &Ti::Vector{ size: rows, .. }) => TypeResolution::Value(Ti::Matrix { columns, rows, width }),
                             (left, right) =>
                                 return Err(ResolveError::IncompatibleOperands(
-                                    format!("{:?}({:?}, {:?})", fun, left, right)
+                                    format!("{fun:?}({left:?}, {right:?})")
                                 )),
                         }
                     },
@@ -746,7 +767,7 @@ impl<'a> ResolveContext<'a> {
                         Ti::Scalar {width,kind} |
                         Ti::Vector {width,kind,size:_} => TypeResolution::Value(Ti::Scalar { kind, width }),
                         ref other => return Err(ResolveError::IncompatibleOperands(
-                                format!("{:?}({:?})", fun, other)
+                                format!("{fun:?}({other:?})")
                             )),
                     },
                     Mf::Normalize |
@@ -772,7 +793,7 @@ impl<'a> ResolveContext<'a> {
                             width,
                         }),
                         ref other => return Err(ResolveError::IncompatibleOperands(
-                            format!("{:?}({:?})", fun, other)
+                            format!("{fun:?}({other:?})")
                         )),
                     },
                     Mf::Inverse => match *res_arg.inner_with(types) {
@@ -786,7 +807,7 @@ impl<'a> ResolveContext<'a> {
                             width,
                         }),
                         ref other => return Err(ResolveError::IncompatibleOperands(
-                            format!("{:?}({:?})", fun, other)
+                            format!("{fun:?}({other:?})")
                         )),
                     },
                     Mf::Determinant => match *res_arg.inner_with(types) {
@@ -795,14 +816,26 @@ impl<'a> ResolveContext<'a> {
                             ..
                         } => TypeResolution::Value(Ti::Scalar { kind: crate::ScalarKind::Float, width }),
                         ref other => return Err(ResolveError::IncompatibleOperands(
-                            format!("{:?}({:?})", fun, other)
+                            format!("{fun:?}({other:?})")
                         )),
                     },
                     // bits
+                    Mf::CountTrailingZeros |
+                    Mf::CountLeadingZeros |
                     Mf::CountOneBits |
                     Mf::ReverseBits |
                     Mf::ExtractBits |
-                    Mf::InsertBits => res_arg.clone(),
+                    Mf::InsertBits |
+                    Mf::FindLsb |
+                    Mf::FindMsb => match *res_arg.inner_with(types)  {
+                        Ti::Scalar { kind: kind @ (crate::ScalarKind::Sint | crate::ScalarKind::Uint), width } =>
+                            TypeResolution::Value(Ti::Scalar { kind, width }),
+                        Ti::Vector { size, kind: kind @ (crate::ScalarKind::Sint | crate::ScalarKind::Uint), width } =>
+                            TypeResolution::Value(Ti::Vector { size, kind, width }),
+                        ref other => return Err(ResolveError::IncompatibleOperands(
+                                format!("{fun:?}({other:?})")
+                            )),
+                    },
                     // data packing
                     Mf::Pack4x8snorm |
                     Mf::Pack4x8unorm |
@@ -821,7 +854,7 @@ impl<'a> ResolveContext<'a> {
                 expr,
                 kind,
                 convert,
-            } => match *past(expr).inner_with(types) {
+            } => match *past(expr)?.inner_with(types) {
                 Ti::Scalar { kind: _, width } => TypeResolution::Value(Ti::Scalar {
                     kind,
                     width: convert.unwrap_or(width),
@@ -835,10 +868,18 @@ impl<'a> ResolveContext<'a> {
                     size,
                     width: convert.unwrap_or(width),
                 }),
+                Ti::Matrix {
+                    columns,
+                    rows,
+                    width,
+                } => TypeResolution::Value(Ti::Matrix {
+                    columns,
+                    rows,
+                    width: convert.unwrap_or(width),
+                }),
                 ref other => {
                     return Err(ResolveError::IncompatibleOperands(format!(
-                        "{:?} as {:?}",
-                        other, kind
+                        "{other:?} as {kind:?}"
                     )))
                 }
             },
@@ -853,6 +894,17 @@ impl<'a> ResolveContext<'a> {
                 kind: crate::ScalarKind::Uint,
                 width: 4,
             }),
+            crate::Expression::RayQueryProceedResult => TypeResolution::Value(Ti::Scalar {
+                kind: crate::ScalarKind::Bool,
+                width: crate::BOOL_WIDTH,
+            }),
+            crate::Expression::RayQueryGetIntersection { .. } => {
+                let result = self
+                    .special_types
+                    .ray_intersection
+                    .ok_or(ResolveError::MissingSpecialType)?;
+                TypeResolution::Handle(result)
+            }
         })
     }
 }
